import numpy as np
from scipy.stats import norm


class MaxAcqQuery:
    def __init__(self, threshold, gp,
                 nr_test, nr_weights,
                 lower=-1.0, upper=1.0,
                 acq="Expected Improvement",
                 **kwargs):
        self.threshold = threshold
        self.gp = gp
        self.nr_test = nr_test
        self.nr_weights = nr_weights
        self.lower = lower
        self.upper = upper
        self.acq = acq

        self.seed = kwargs.get('seed', None)
        self.kappa = kwargs.get('kappa', 2.576)
        self.beta = kwargs.get('beta', 1.2)
        self.X = kwargs.get('X', None)

        self.rng = np.random.default_rng(self.seed)

    def query(self):
        X_test = self.rng.uniform(self.lower, self.upper, (self.nr_test, self.nr_weights))
        max_acq = 0

        if self.acq == "Expected Improvement":
            if self.X is None:
                raise ValueError
            y_hat = self.gp.predict(self.X)
            best_y = max(y_hat)
            mu, sigma = self.gp.predict(X_test, return_std=True)
            z = (mu - best_y - self.kappa) / sigma
            ei = (mu - best_y - self.kappa) * norm.cdf(z) + sigma * norm.pdf(z)
            max_acq = np.max(ei)

        if self.acq == "Probability of Improvement":
            if self.X is None:
                raise ValueError
            y_hat = self.gp.predict(self.X)
            best_y = max(y_hat)
            mu, sigma = self.gp.predict(X_test, return_std=True)
            z = (mu - best_y - self.kappa) / sigma
            pi = norm.cdf(z)
            max_acq = np.max(pi)

        if self.acq == "Upper Confidence Bound":
            mu, sigma = self.gp.predict(X_test, return_std=True)
            cb = mu + self.beta * sigma
            max_acq = np.max(cb)

        return max_acq > self.threshold
