from active_bo_msgs.msg import ActiveRLRequest
from active_bo_msgs.msg import ActiveRLResponse
from active_bo_msgs.msg import ActiveRLEvalRequest
from active_bo_msgs.msg import ActiveRLEvalResponse

from active_bo_msgs.msg import ImageFeedback

import rclpy
from rclpy.node import Node

from rclpy.callback_groups import ReentrantCallbackGroup

from active_bo_ros.ReinforcementLearning.ContinuousMountainCar import Continuous_MountainCarEnv
from active_bo_ros.ReinforcementLearning.CartPole import CartPoleEnv
from active_bo_ros.ReinforcementLearning.Pendulum import PendulumEnv
from active_bo_ros.ReinforcementLearning.Acrobot import AcrobotEnv


import numpy as np
import time
import copy


class ActiveRL(Node):
    def __init__(self):
        super().__init__('active_rl_service')
        rl_callback_group = ReentrantCallbackGroup()
        topic_callback_group = ReentrantCallbackGroup()
        mainloop_callback_group = ReentrantCallbackGroup()

        # Active Reinforcement Learning Publisher, Subscriber and Message attributes
        self.active_rl_pub = self.create_publisher(ActiveRLResponse,
                                                   'active_rl_response',
                                                   1, callback_group=rl_callback_group)
        self.active_rl_sub = self.create_subscription(ActiveRLRequest,
                                                      'active_rl_request',
                                                      self.active_rl_callback,
                                                      1, callback_group=rl_callback_group)

        self.rl_env = None
        self.rl_seed = None
        self.rl_policy = None
        self.rl_weights = None
        self.rl_reward = 0.0
        self.rl_step = 0

        # Image publisher to publish the rgb array from the gym environment
        self.image_pub = self.create_publisher(ImageFeedback,
                                               'rl_feedback',
                                               1, callback_group=topic_callback_group)

        # Active RL Evaluation Publisher, Subscriber and Message attributes
        self.eval_pub = self.create_publisher(ActiveRLEvalRequest,
                                              'active_rl_eval_request',
                                              1,
                                              callback_group=topic_callback_group)
        self.eval_sub = self.create_subscription(ActiveRLEvalResponse,
                                                 'active_rl_eval_response',
                                                 self.active_rl_eval_callback,
                                                 1,
                                                 callback_group=topic_callback_group)

        self.eval_response_received = False
        self.eval_policy = None
        self.eval_weights = None
        self.overwrite_weight = None

        # RL Environments
        self.env = None

        # State Machine Variables
        self.best_pol_shown = False
        self.policy_sent = False
        self.rl_pending = False
        self.interactive_run = 0
        self.display_run = False

        # Main loop timer object
        self.mainloop_timer_period = 0.1
        self.mainloop = self.create_timer(self.mainloop_timer_period,
                                          self.mainloop_callback,
                                          callback_group=mainloop_callback_group)

    def reset_rl_request(self):
        self.rl_env = None
        self.rl_seed = None
        self.rl_policy = None
        self.rl_weights = None
        self.interactive_run = 0
        self.display_run = False

    def active_rl_callback(self, msg):
        self.rl_env = msg.env
        self.rl_seed = msg.seed
        self.display_run = msg.display_run
        self.rl_policy = np.array(msg.policy, dtype=np.float64)
        self.rl_weights = msg.weights
        self.interactive_run = msg.interactive_run

        if self.rl_env == "Mountain Car":
            self.env = Continuous_MountainCarEnv(render_mode="rgb_array")
        elif self.rl_env == "Cartpole":
            self.env = CartPoleEnv(render_mode="rgb_array")
        elif self.rl_env == "Acrobot":
            self.env = AcrobotEnv(render_mode="rgb_array")
        elif self.rl_env == "Pendulum":
            self.env = PendulumEnv(render_mode="rgb_array")
        else:
            raise NotImplementedError

        # self.get_logger().info('Active RL: Called!')
        self.env.reset(seed=self.rl_seed)
        self.rl_pending = True
        self.policy_sent = False
        self.rl_step = 0

    def reset_eval_request(self):
        self.eval_policy = None
        self.eval_weights = None

    def active_rl_eval_callback(self, msg):
        self.eval_policy = np.array(msg.policy, dtype=np.float64)
        self.eval_weights = msg.weights
        self.overwrite_weight = msg.overwrite_weight

        self.get_logger().info('Active RL Eval: Responded!')
        self.env.reset(seed=self.rl_seed)
        self.eval_response_received = True

    def next_image(self, policy, display_run):
        action = policy[self.rl_step]
        action_clipped = action.clip(min=-1.0, max=1.0)
        output = self.env.step(action_clipped.astype(np.float64))

        self.rl_reward += output[1]
        done = output[2]
        self.rl_step += 1

        if display_run:
            rgb_array = self.env.render()
            rgb_shape = rgb_array.shape

            red = rgb_array[:, :, 0].flatten().tolist()
            green = rgb_array[:, :, 1].flatten().tolist()
            blue = rgb_array[:, :, 2].flatten().tolist()

            feedback_msg = ImageFeedback()

            feedback_msg.height = rgb_shape[0]
            feedback_msg.width = rgb_shape[1]
            feedback_msg.current_time = self.rl_step
            feedback_msg.red = red
            feedback_msg.green = green
            feedback_msg.blue = blue

            self.image_pub.publish(feedback_msg)

        if not done and self.rl_step == len(policy):
            done = True

        return done

    def complete_run(self, policy):
        env_reward = 0.0
        step_count = 0

        self.env.reset(seed=self.rl_seed)

        for i in range(len(policy)):
            action = policy[i]
            action_clipped = action.clip(min=-1.0, max=1.0)
            output = self.env.step(action_clipped.astype(np.float64))

            env_reward += output[1]
            done = output[2]
            step_count += 1

            if done:
                break

        self.env.reset(seed=self.rl_seed)
        return env_reward, step_count

    def mainloop_callback(self):
        if self.rl_pending:
            if self.interactive_run == 0:
                if not self.best_pol_shown:
                    if not self.policy_sent:
                        self.rl_step = 0
                        self.rl_reward = 0.0
                        self.env.reset(seed=self.rl_seed)

                        eval_request = ActiveRLEvalRequest()
                        eval_request.policy = self.rl_policy.tolist()
                        eval_request.weights = self.rl_weights

                        self.eval_pub.publish(eval_request)
                        self.get_logger().info('Active RL: Called!')
                        self.get_logger().info('Active RL: Waiting for Eval!')

                        self.policy_sent = True

                    done = self.next_image(self.rl_policy, self.display_run)

                    if done:
                        self.best_pol_shown = True
                        self.rl_step = 0
                        self.rl_reward = 0.0

                elif self.best_pol_shown:
                    if not self.eval_response_received:
                        pass

                    if self.eval_response_received:
                        done = self.next_image(self.eval_policy, self.display_run)

                        if done:
                            rl_response = ActiveRLResponse()
                            rl_response.weights = self.eval_weights
                            rl_response.reward = self.rl_reward
                            rl_response.final_step = self.rl_step
                            rl_response.overwrite_weight = self.overwrite_weight

                            self.active_rl_pub.publish(rl_response)

                            self.env.reset(seed=self.rl_seed)

                            # reset flags and attributes
                            self.reset_eval_request()
                            self.reset_rl_request()

                            self.rl_step = 0
                            self.rl_reward = 0.0

                            self.best_pol_shown = False
                            self.eval_response_received = False
                            self.rl_pending = False

            elif self.interactive_run == 1:
                if not self.policy_sent:
                    self.rl_step = 0
                    self.rl_reward = 0.0
                    self.env.reset(seed=self.rl_seed)

                    eval_request = ActiveRLEvalRequest()
                    eval_request.policy = self.rl_policy.tolist()
                    eval_request.weights = self.rl_weights

                    self.eval_pub.publish(eval_request)
                    self.get_logger().info('Active RL: Called!')
                    self.get_logger().info('Active RL: Waiting for Eval!')

                    self.policy_sent = True

                done = self.next_image(self.rl_policy, self.display_run)

                if done:
                    self.rl_step = 0
                    self.rl_reward = 0.0
                    self.rl_pending = False

            elif self.interactive_run == 2:
                env_reward, step_count = self.complete_run(self.rl_policy)

                rl_response = ActiveRLResponse()
                rl_response.weights = self.rl_weights
                rl_response.reward = env_reward
                rl_response.final_step = step_count
                if self.overwrite_weight is None:
                    overwrite_weight = [False] * len(self.rl_weights)
                else:
                    overwrite_weight = self.overwrite_weight

                rl_response.overwrite_weight = overwrite_weight

                self.active_rl_pub.publish(rl_response)

                self.reset_rl_request()
                self.rl_pending = False


def main(args=None):
    rclpy.init(args=args)

    active_rl_service = ActiveRL()

    rclpy.spin(active_rl_service)

    rclpy.shutdown()


if __name__ == '__main__':
    main()
